import os
from django.http.response import Http404
from django.conf import settings
from django.utils._os import safe_join
TERM = "terminal" in settings.INSTALLED_APPS
if TERM is True and "django_downloadview" in settings.INSTALLED_APPS is True:
    from django_downloadview import PathDownloadView


class DexDlView(PathDownloadView):
    app_dir = os.path.dirname(os.path.abspath(__file__))
    project_dir = os.path.dirname(app_dir)
    rpath = safe_join(project_dir, 'replica.sqlite3')

    def dispatch(self, request, *args, **kwargs):
        if not self.request.user.is_superuser:
            raise Http404
        return super(DexDlView, self).dispatch(request, *args, **kwargs)

    def get_path(self):
        return self.rpath


class DexMediaView(PathDownloadView):
    app_dir = os.path.dirname(os.path.abspath(__file__))
    project_dir = os.path.dirname(app_dir)
    rpath = safe_join(project_dir, 'media.zip')

    def dispatch(self, request, *args, **kwargs):
        if not self.request.user.is_superuser:
            raise Http404
        return super(DexMediaView, self).dispatch(request, *args, **kwargs)

    def get_path(self):
        return self.rpath



from goerr import err
from django.conf import settings
from django.db import IntegrityError
from ...conf import PREFETCH


class DjangoDb():

    def _get_django_db(self, dbname):
        dbs = getattr(settings, "DATABASES")
        for name in dbs:
            db = dbs[name]
            if name == dbname:
                return db
        return

    def _queryset(self, model, dbsource, prefetch=True):
        modelname = model.__name__
        if modelname in PREFETCH and prefetch is True:
            qs = model.objects.using(dbsource).all().prefetch_related(
                *PREFETCH[modelname])
        else:
            qs = model.objects.using(dbsource).all()
        return qs

    def _save_instance(self, model, instance, dbdest):
        try:
            self._disable_auto_now_fields(model)
            instance.save(using=dbdest, force_insert=True)
        except IntegrityError as e:
            msg = "ERROR inserting", instance, "- ", model.__name__
            err.new(e, self._save_instance, msg)
            raise(e)
        return

    def _disable_auto_now_fields(self, *models):
        for model in models:
            for field in model._meta.local_fields:
                if hasattr(field, 'auto_now'):
                    field.auto_now = False
                if hasattr(field, 'auto_now_add'):
                    field.auto_now_add = False



from influxdb import InfluxDBClient


CLI = None
POINTS = []


def init(db):
    global CLI
    CLI = InfluxDBClient(
        db.conf["host"],
        db.conf["port"],
        db.conf["user"],
        db.conf["password"],
        db.conf["db"],
    )


def write(point, force_save=False):
    global CLI
    global POINTS
    POINTS.append(point)
    if len(POINTS) < 100 and force_save is False:
        return
    try:
        CLI.write_points(POINTS, batch_size=1000, time_precision="s")
        POINTS = []
    except Exception as err:
        raise err
    return



from django.conf import settings


EXCLUDE = getattr(settings, 'DEX_EXCLUDE', ["filebrowser", "admin"])
PREFETCH = getattr(settings, "DEX_PREFETCH", {})
MEASUREMENT = getattr(settings, 'DEX_MEASUREMENT', "django_model")
TIME_FIELD = getattr(settings, 'DEX_DEFAULT_TIME_FIELD', "date")
TIME_FIELDS = getattr(settings, 'DEX_TIME_FIELDS', {
                      "User": "date_joined", "MEvent": "date_posted"})
SITE_SLUG = getattr(settings, "SITE_SLUG")
COMMAND_CHANNEL = getattr(
    settings, "DEX_COMMAND_CHANNEL", "$" + SITE_SLUG + "_command")

DBS = getattr(settings, "DEX_DBS", {})



from __future__ import print_function
import time
from goerr import err
from django.conf import settings
from django.contrib.auth.models import Permission
from django.contrib.sites.models import Site
from introspection.inspector import inspect
#from ..db import DexDb
from ..db.django import DjangoDb
from ..conf import EXCLUDE


class Exporter(DjangoDb):

    def __init__(self, dbname=None):
        self.errors = []
        if dbname is None:
            return
        self.err = None
        """
        self.db = DexDb(dbname)
        if self.db.err is not None:
            err = "Dex exporter initialization error:\n" + self.db.err
            self.err = err
        """

    def clone(self, dbsource, dbdest, applist=None, verbosity=1):
        global EXCLUDE
        st = time.time()
        if applist is None:
            applist = settings.INSTALLED_APPS
        source = self._get_django_db(dbsource)
        if source is None:
            msg = "Database", dbsource, "not found"
            err.new(self.clone, msg)
            return
        dest = self._get_django_db(dbdest)
        if dest is None:
            msg = "Database", dbdest, "not found"
            err.new(self.clone, msg)
            return
        if verbosity > 0:
            print("Cloning database", dbsource, "in database", dbdest)
        models = {}
        for appname in applist:
            appstr = appname.split('.')[-1]
            if appstr in EXCLUDE or appstr.startswith("django.") \
                    or appstr == "django.contrib.contenttypes":
                continue
            # remove all remaining dots to keep just the app name
            try:
                s = appname.split(".")
                appname = s[len(s) - 1]
            except:
                pass
            models[appname] = inspect.models(appname)
        num_models = 0
        num_instances = 0
        stats = {}
        appname = "contenttypes"
        if verbosity > 0:
            print("Found", str(len(applist)), "applications")
        for appname in models:
            if appname == "contenttypes" or appname == "sessions":
                continue
            else:
                if len(models[appname]) > 0:
                    print("# Processing app", appname)
                    stats[appname] = {}
                    stats[appname]["num_models"] = 0
                    num_models = 0
                    for model in models[appname]:
                        if model == Permission or model == Site:
                            continue
                        num_models += 1
                        stats, num_instances = self.clone_model(model, dbsource, dbdest,
                                                                num_instances, appname,
                                                                models, num_models, stats, verbosity)
                        err.fatal()
        self.stats(models, stats, num_models, num_instances, st)

    def clone_model(self, model, dbsource, dbdest, num_instances,
                    appname, models, num_models, stats, verbosity):
        qs = self._queryset(model, dbsource, False)
        if verbosity > 0:
            print("- Model", model.__name__, ":", qs.count(), "objects found")
        num_model_instances = 0
        for instance in qs:
            print(num_instances, appname, "-",
                  model.__name__, num_model_instances)
            num_instances += 1
            num_model_instances += 1
            try:
                self._save_instance(model, instance, dbdest)
            except Exception as e:
                err.new(e, self.clone_model,
                        "Can not clone model " + model.__name__)
                return
        num_models += 1
        stats[appname]["num_models"] = len(models[appname])
        stats[appname][model.__name__] = num_model_instances
        return stats, num_instances

    def stats(self, models, stats, num_models, num_instances, st):
        num_apps = len(models)
        elapsed_time = round(time.time() - st, 2)
        for appname in stats:
            if len(appname) > 1 and stats[appname]["num_models"] > 0:
                print("###", appname, ":",
                      stats[appname]["num_models"], "models")

            for m in stats[appname]:
                ni = stats[appname][m]
                if ni > 0 and m != "num_models":
                    print("|--", m, ":", ni, "instances")
                elif ni == 0:
                    print("|xx", m, ": no instances")

        print("[OK] Saved", num_instances, "instances from", num_models,
              "models from", num_apps, "apps in", elapsed_time, "s")
        if len(self.errors) > 0:
            print("ERRORS")
            for err in self.errors:
                print(err)



from terminal.commands import Command, rprint
from django.core.management import call_command
from django.utils.six import StringIO
from dex.exporter import Exporter
from django.urls.base import reverse


def replicatedb(request, cmd_args):
    rprint("Migrating replica, please wait ...")
    ex = Exporter()
    out = StringIO()
    call_command("migrate", "--database=replica", "--no-color", stdout=out)
    res = out.getvalue()
    endres = res.split("\n")
    for r in endres:
        rprint(r)
    ex.clone("default", "replica", None)
    url = reverse("dex-dl")
    rprint('Ok: <a href="' + url + '">download the replica</a>')
    return None


c1 = Command("replicatedb", replicatedb, "Replicate the database")

COMMANDS = [c1]



from __future__ import print_function
from goerr import err
from django.core.management.base import BaseCommand
from django.core.management import call_command
from dex.export import Exporter


class Command(BaseCommand):
    help = 'Clone a database: python manage.py clonedb sourcedb_name destinationdb_name'

    def add_arguments(self, parser):
        parser.add_argument('source', type=str)
        parser.add_argument('dest', type=str)
        parser.add_argument('-apps',
                            dest='applist',
                            default=None,
                            help='Applications to clone: ex: app1,app2,app3',
                            )
        parser.add_argument('-m',
                            dest='migrate',
                            action='store_true',
                            default=False,
                            help='Migrate the destination database',
                            )
        parser.add_argument('-verb',
                            dest='verbosity',
                            default=1,
                            help='Set verbosity',
                            )

    def handle(self, *args, **options):
        try:
            ex = Exporter()
        except Exception as e:
            err.new(e, self.handle, "Can not initialize exporter")
        source = options["source"]
        dest = options["dest"]
        applist = options["applist"]
        if options["migrate"] is not False:
            try:
                call_command("migrate", "--database=" + dest)
            except Exception as e:
                err.new(e, self.handle,
                        "Can not migrate destination database " + dest)
        if applist is not None:
            applist = str.split(options["applist"], ",")
        try:
            ex.clone(source, dest, applist=applist,
                     verbosity=int(options["verbosity"]))
        except Exception as e:
            err.new(e, self.handle,
                    "Can not clone database")
        if err.exists:
            err.throw()



import datetime
from django.core.exceptions import FieldDoesNotExist
from dex.conf import TIME_FIELDS


def serialize_protocol(model, instance, data, measurement, time_field, enable_text_field):
    res = _format_data(model, instance, data, measurement,
                       time_field, TIME_FIELDS, enable_text_field)
    return res


def _to_date(val):
    return val.strftime('%Y-%m-%dT%H:%M:%SZ')


def _format_data(model, instance, data, measurement, time_field, time_fields, enable_text_field):
    modelname = model.__name__
    tags = {"modelname": modelname}
    t = int(datetime.datetime.now().timestamp())
    timefield = None
    is_timefield = False
    if modelname in time_fields:
        timefield = time_fields[modelname]
    for field in data:
        try:
            ftype = model._meta.get_field(field).get_internal_type
        except FieldDoesNotExist:
            ftype = "unknown"
        val = data[field]
        if 'DateTimeField' in str(ftype) and timefield is None:
            val = getattr(instance, field)
            if val is not None:
                tv = int(val.timestamp())
                tags[field] = tv
        if is_timefield is False:
            if enable_text_field is False:
                if not "TextField" in str(ftype):
                    tags[field] = val
                else:
                    tags[field] = val
        if field == time_field or (timefield is not None and field == timefield):
            val = getattr(instance, field)
            if val is not None:
                tv = int(val.timestamp())
                t = _to_date(val)

    fields = {"num": 1}
    data = {
        "measurement": measurement,
        "tags": tags,
        "fields": fields,
        "time": t,
    }
    return data
